# Copyright (c) Qotto, 2021

"""
Record abstract base class
"""

from datetime import datetime, timezone
from typing import Any, Dict, Union, Optional
from uuid import UUID, uuid4

from semver import VersionInfo
from urnparse import URN8141, InvalidURNFormatError

from .record_errors import RecordAttributeTypeError, RecordAttributeValueError

__all__ = [
    'Record',
]


class Record:
    """
    Common Record interface

    This class is a base for Event, Request, and Response. It should not be instantiated.
    """

    def __init__(
        self,
        protocol_version: str,
        schema: str,
        version: str,
        source: str,
        uuid: str = None,
        correlation_id: str = None,
        partition_key: str = None,
        date: datetime = None,
        context: Dict[str, Dict[str, Any]] = None,
        data: Dict[str, Any] = None

    ) -> None:
        """
        Initialize Record attributes

        :param protocol_version: SemVer version of the Eventy protocol
        :param schema: URN of the record schema
        :param version: SemVer version of the record schema
        :param source: URN of the record emitter
        :param uuid: UUID v4 of the record. Default: auto-generated
        :param correlation_id: Trace / correlation ID for trace_id. Default: auto-generated
        :param partition_key: Partition or database key. Default: None
        :param date: Date of the event. Default: now
        :param context: Keys should be URN of a service using the context
        :param data: Domain data

        :raises RecordAttributeError:
        """
        self.protocol_version = protocol_version
        self.schema = schema
        self.version = version
        self.source = source
        self.uuid = uuid  # type: ignore
        self.correlation_id = correlation_id  # type: ignore
        self.partition_key = partition_key  # type: ignore
        self.date = date  # type: ignore
        self.context = context  # type: ignore
        self.data = data  # type: ignore

    @property
    def type(self) -> str:
        """
        Type of the record: EVENT | REQUEST | RESPONSE
        """
        raise NotImplementedError

    @property
    def protocol_version(self) -> str:
        """
        Eventy protocol version (SemVer)
        """
        return self._protocol_version

    @protocol_version.setter
    def protocol_version(self, protocol_version: str):
        try:
            VersionInfo.parse(protocol_version)
        except ValueError:
            raise RecordAttributeValueError('protocol_version', protocol_version, 'Not in SemVer format')
        self._protocol_version = protocol_version

    @property
    def schema(self) -> str:
        """
        Record schema (URN)
        """
        return self._schema

    @schema.setter
    def schema(self, schema: str):
        try:
            URN8141.from_string(schema)
        except InvalidURNFormatError:
            raise RecordAttributeValueError('schema', schema, 'Not in URN format')
        self._schema = schema

    @property
    def version(self) -> str:
        """
        Record schema version (SemVer)
        """
        return self._version

    @version.setter
    def version(self, version: str):
        try:
            VersionInfo.parse(version)
        except ValueError:
            raise RecordAttributeValueError('version', version, 'Not in SemVer format')
        self._version = version

    @property
    def source(self) -> str:
        """
        Record source (URN)
        """
        return self._source

    @source.setter
    def source(self, source: str):
        try:
            URN8141.from_string(source)
        except InvalidURNFormatError:
            raise RecordAttributeValueError('source', source, 'Not in URN format')
        self._source = source

    @property
    def uuid(self) -> str:
        """
        Record unique identifier (UUID v4)
        """
        return self._uuid

    @uuid.setter
    def uuid(self, uuid: str):
        if uuid is None:
            uuid = str(uuid4())
        try:
            UUID(uuid)
        except TypeError as e:
            raise RecordAttributeValueError('uuid', uuid, str(e)) from e
        except ValueError as e:
            raise RecordAttributeValueError('uuid', uuid, str(e)) from e
        self._uuid = uuid

    @property
    def correlation_id(self) -> str:
        """
        Record correlation_id to be propagated
        """
        return self._trace_id

    @correlation_id.setter
    def correlation_id(self, trace_id: str):
        if trace_id is None:
            trace_id = f'{uuid4()}:{int(datetime.now(timezone.utc).timestamp() * 1000)}'
        if isinstance(trace_id, str):
            self._trace_id = trace_id
        else:
            raise RecordAttributeTypeError('trace_id', str, trace_id)

    @property
    def partition_key(self) -> Optional[str]:
        """
        Key for partitioned states, e.g. Kafka topics partitions
        """
        return self._key

    @partition_key.setter
    def partition_key(self, key: str):
        if key is None or isinstance(key, str):
            self._key = key
        else:
            raise RecordAttributeTypeError('partition_key', str, key)

    @property
    def date(self) -> datetime:
        """
        Date of the event (unrelated to event production)
        """
        return self._date

    @date.setter
    def date(self, date: Union[datetime, int, str]):
        """
        Set the date, keeping a millisecond precision.

        :param date: record date, as a datetime, or iso str, or timestamp int
        """
        if date is None:
            date = datetime.now(timezone.utc)
        elif isinstance(date, str):
            try:
                date = datetime.fromisoformat(date)
            except Exception:
                raise RecordAttributeValueError('date', date, 'Not in iso format')
        elif isinstance(date, int):
            try:
                date = datetime.fromtimestamp(date / 1000, timezone.utc)
            except Exception:
                raise RecordAttributeValueError('date', date, 'Not a ms timestamp')

        if isinstance(date, datetime):
            ts = int(date.timestamp() * 1000) / 1000
            date = datetime.fromtimestamp(ts, timezone.utc)
            self._date = date
        else:
            raise RecordAttributeTypeError('date', datetime, date)

    @property
    def context(self) -> Dict[str, Dict[str, Any]]:
        """
        Execution context to be propagated
        """
        return self._context

    @context.setter
    def context(self, context: Dict[str, Dict[str, Any]]):
        if context is None:
            context = dict()
        if not isinstance(context, Dict):
            raise RecordAttributeTypeError('context', Dict, context)
        for key, val in context.items():
            if not isinstance(key, str):
                raise RecordAttributeTypeError(f'context key', str, key)
            if not isinstance(val, Dict):
                raise RecordAttributeTypeError(f'value for key {key}', 'Dict', val)
        self._context = context

    @property
    def data(self) -> Dict[str, Any]:
        """
        Actual record payload
        """
        return self._data

    @data.setter
    def data(self, data):
        self._data = dict()
        if data is None:
            data = dict()
        for key in data:
            if not isinstance(key, str):
                raise RecordAttributeTypeError('data key', 'str', key)
        self._data = data

    def __repr__(self):
        return f"{self.type} {self.schema} v{self.version} uuid={self.uuid} trace_id={self.correlation_id} key={self.partition_key}"

    def _debug_str(self):
        return self.__repr__() + f"\n      CONTEXT={self.context}\n      DATA={self.data}"



# Copyright (c) Qotto, 2021

"""
Google Kubernetes Engine (GKE) logging handler

>>> import sys
>>> import logging
>>> logger = logging.getLogger('my_logger')
>>> logger.setLevel('DEBUG')
>>> handler = GkeHandler(stream=sys.stdout)
>>> logger.addHandler(handler)
>>> logger.info('test')
{\
"timestamp": {"seconds": ..., "nanos": ...}, "severity": "INFO", "message": "test", \
"file": "<doctest eventy.logging.gke_handler[6]>", "line": 1, "module": "<doctest eventy.logging", \
"function": "<module>", "logger_name": "my_logger", "thread": ...\
}
"""

import json
import logging
import math

from .info_providers import correlation_id_provider, request_id_provider

__all__ = [
    'GkeHandler',
]


class GkeHandler(logging.StreamHandler):
    """
    GKE logging handler formats log messages to Google Kubernetes Engine JSON format
    """

    def __init__(self, stream=None, level='DEBUG') -> None:
        """
        Initialize the handler

        :param stream: output stream, default None, propagated to StreamHandler.__init__
        :param level: log level, default 'DEBUG', should be less than Logger.setLevel()
        """
        super().__init__(stream=stream)
        self.addFilter(correlation_id_provider)
        self.addFilter(request_id_provider)
        self.setLevel(level=level)

    def format(self, record: logging.LogRecord) -> str:
        """
        Override default formatting and create a JSON in GKE format

        :param record: log record to format
        :return: formatted JSON string
        """
        message = super().format(record)

        subsecond, second = math.modf(record.created)
        payload = {
            'timestamp': {
                'seconds': int(second),
                'nanos': int(subsecond * 1e9)
            },
            'severity': record.levelname,
            'message': message,
            'file': record.pathname,
            'line': record.lineno,
            'module': record.module,
            'function': record.funcName,
            'logger_name': record.name,
            'thread': record.thread,
            'correlation_id': record.correlation_id,  # type: ignore
            'request_id': record.request_id  # type: ignore
        }
        if hasattr(record, 'correlation_id'):
            payload['correlation_id'] = getattr(record, 'correlation_id')
        if hasattr(record, 'request_id'):
            payload['request_id'] = getattr(record, 'request_id')
        return json.dumps(payload)



# Copyright (c) Qotto, 2021

import json
from datetime import datetime
from typing import Union, Any, Dict

from eventy.serialization import RecordSerializer, SerializationError
from eventy.serialization.serialization_errors import UnknownRecordTypeError
from eventy.record import Record, Event, Request, Response


class JsonSerializer(RecordSerializer):

    def encode(self, record: Record) -> bytes:
        record_dict: Dict[str, Any] = {
            'type': record.type,
            'protocol_version': record.protocol_version,
            'schema': record.schema,
            'version': record.version,
            'source': record.source,
            'uuid': record.uuid,
            'correlation_id': record.correlation_id,
            'partition_key': record.partition_key,
            'date_timestamp': int(record.date.timestamp() * 1000),
            'date_iso8601': record.date.isoformat(),
        }
        if isinstance(record, (Request, Response)):
            record_dict.update(
                {
                    'destination': str(record.destination),
                }
            )
        if isinstance(record, Response):
            record_dict.update(
                {
                    'request_uuid': record.request_uuid,
                    'ok': record.ok,
                    'error_code': record.error_code,
                    'error_message': record.error_message,
                }
            )
        record_dict.update(
            {
                'context': record.context,
                'data': record.data
            }
        )
        return bytes(json.dumps(record_dict), encoding='utf-8')

    def decode(self, encoded: bytes) -> Union[Event, Request, Response]:
        try:
            record_json = json.loads(encoded)
            record_type = record_json.pop('type')
            date = datetime.fromisoformat(record_json.pop('date_iso8601'))
            record_json.pop('date_timestamp')
            record_json['date'] = date
            record: Union[Event, Request, Response]
            if record_type == 'EVENT':
                record = Event(**record_json)
            elif record_type == 'REQUEST':
                record = Request(**record_json)
            elif record_type == 'RESPONSE':
                record = Response(**record_json)
            else:
                raise UnknownRecordTypeError(record_type)
            return record
        except Exception as e:
            raise SerializationError from e



# Copyright (c) Qotto, 2021

"""
Eventy configuration
"""

__all__ = [
    'SERVICE_NAME',
]

SERVICE_NAME: str = '-'
"""
Do not update with:
>>> from eventy.config import SERVICE_NAME
>>> SERVICE_NAME = 'my_service'

If you want to update the value you need to do:
>>> from eventy import config
>>> config.SERVICE_NAME = 'my_service'
"""



# Copyright (c) Qotto, 2021

"""
<<<<<<< HEAD
Provides information to log records

Logging filters to inject trace ids in LogRecord instances

The python logging API allows to add attributes to log records using filters
"""

from logging import LogRecord

from eventy.trace_id import correlation_id_var, request_id_var

__all__ = [
    'correlation_id_provider',
    'request_id_provider',
]


def correlation_id_provider(log_record: LogRecord) -> bool:
    """
    Add ``correlation_id`` field in the log record

    :param log_record: log record to annotate
    :return: True (not really a filter)
    """
    setattr(log_record, 'correlation_id', correlation_id_var.get())
    return True


def request_id_provider(log_record: LogRecord) -> bool:
    """
    Add ``request_id`` field in the log record

    :param log_record: log record to annotate
    :return: True (not really a filter)
    """
    setattr(log_record, 'request_id', request_id_var.get())
    return True



# Copyright (c) Qotto, 2021

"""
Celery integration utilities

Utility functions to integrate the eventy protocol in celery apps

You need to install celery optional dependencies ("pip install eventy[celery]")
"""

import logging
from contextvars import copy_context
from functools import wraps
from typing import Callable, List, Dict, Any

from celery import shared_task as celery_shared_task
from celery.execute import send_task as celery_send_task
from celery.result import AsyncResult

from eventy.trace_id import correlation_id_var, request_id_var
from eventy.trace_id.generator import gen_trace_id

logger = logging.getLogger(__name__)

__all__ = [
    'send_task',
    'shared_task',
]


def send_task(name: str, args: List = None, kwargs: Dict = None, **options) -> AsyncResult:
    """
    Modified version of celery.execute.send_task adding context in task kwargs

    options ar propagated as celery send_task named arguments and keyword arguments

    Usage::

        from eventy.integration.celery import send_task

        send_task(
            'service.task.name', [],
            {
                'param': 'value',
            }
        )

    If a correlation_id is defined in the current context, it is equivalent to::

        from celery.execute import send_task
        from eventy.trace_id import correlation_id_var

        send_task(
            'service.task.name', [],
            {
                'correlation_id': correlation_id_var.get(),
                'param': 'value',
            }
        )
    """
    if args is None:
        args = []
    if kwargs is None:
        kwargs = {}

    if correlation_id_var.get():
        kwargs.update(correlation_id=correlation_id_var.get())

    return celery_send_task(name, args, kwargs, **options)


def shared_task(_func: Callable = None, **kwargs) -> Callable:
    """
    Modified version of celery.shared_task adding traces in the task keyword arguments

    ``correlation_id`` is fetched from context, or generated. ``request_id`` is always generated

    Usage::

        from eventy.integration.celery import shared_task

    Instead of::

        from celery import shared_task

    And then::

        @shared_task(options...)
        def my_task(param):
            ...


    :param _func: if used as a decorator with no args, decorated function will be first arg
    :param kwargs: options are propagated to celery's shared_task
    :return: new version of shared_task
    """

    def _shared_task_decorator(func: Callable) -> Callable:
        def _run_and_extract_trace_id(*func_args, **func_kwargs) -> Any:
            if 'correlation_id' in func_kwargs:
                correlation_id_var.set(func_kwargs.pop('correlation_id'))
            else:
                correlation_id_var.set(gen_trace_id(func))
            request_id_var.set(gen_trace_id(func))
            return func(*func_args, **func_kwargs)

        @wraps(func)
        def _run_inside_context(*func_args, **func_kwargs) -> Any:
            context = copy_context()
            return context.run(_run_and_extract_trace_id, *func_args, **func_kwargs)

        return celery_shared_task(**kwargs)(_run_inside_context)

    if _func:
        return _shared_task_decorator(_func)
    else:
        return _shared_task_decorator



# Copyright (c) Qotto, 2021

"""
Kafka admin client
"""

from typing import Dict

from confluent_kafka.admin import AdminClient, ClusterMetadata, TopicMetadata, NewTopic

from .kafka_errors import KafkaTopicCreationError, KafkaTopicDeletionError, KafkaError

__all__ = [
    'KafkaAdmin',
]


class KafkaAdmin:
    """
    Wrapper around a Confluent Kafka AdminClient
    """

    def __init__(
        self,
        admin_config: Dict[str, str],
    ) -> None:
        self._kafka_admin = AdminClient(admin_config)

    def create_topic(self, topic: str, partitions: int = 1, replications: int = 1) -> None:
        """
        Create a topic

        :param topic: topic name to create
        :param partitions: number of partitions, default 1
        :param replications: number of replicas, default 1

        :raise KafkaTopicCreationError: Could not create the topic (e.g. because it already exists)
        """
        try:
            future_topics = self._kafka_admin.create_topics([NewTopic(topic, partitions, replications)])
            for future_topic in future_topics.values():
                future_topic.result()
        except Exception as e:
            raise KafkaTopicCreationError(f"Could not create topic {topic}") from e

    def delete_topic(self, topic: str) -> None:
        """
        Delete a topic

        :param topic: topic name to delete

        :raise KafkaTopicDeletionError: Could not delete the topic (e.g. because it does not exist)
        """
        try:
            future_topics = self._kafka_admin.delete_topics([topic])
            for future_topic in future_topics.values():
                future_topic.result()
        except Exception as e:
            raise KafkaTopicDeletionError(f"Could not delete topic {topic}") from e

    def list_topic(self) -> list[str]:
        """
        List all topics in the broker

        :return: The list of all topic names

        :raise KafkaError: The broker did not answer
        """
        try:
            cluster_meta: ClusterMetadata = self._kafka_admin.list_topics()
            topic_meta: TopicMetadata
            return sorted(cluster_meta.topics.keys())
        except Exception as e:
            raise KafkaError from e



# Copyright (c) Qotto, 2021

"""
Trace IDs generation utilities
"""

from secrets import token_urlsafe
from typing import Callable, Union

import eventy.config

__all__ = [
    'gen_trace_id',
]


def gen_trace_id(func_or_str: Union[Callable, str]) -> str:
    """
    Generate a trace id from a function or a string

    Uses :obj:`eventy.config.SERVICE_NAME`, given str or func __name__ or str() of param, and a random string
    :param func_or_str: function where the trace id is defined, or specific string
    :return: a new trace id "SERVICE_NAME:func_or_str:RANDOM"
    """
    if isinstance(func_or_str, str):
        return f'{eventy.config.SERVICE_NAME}:{func_or_str}:{token_urlsafe(8)}'
    elif hasattr(func_or_str, '__name__'):
        return gen_trace_id(func_or_str.__name__)
    else:
        return gen_trace_id(str(func_or_str))



# Copyright (c) Qotto, 2021

"""
Record serialization package

Define the interfaces for record serialization.
"""

from .record_serializer import RecordSerializer, DestinationRecordSerializer, ContextRecordSerializer
from .serialization_errors import SerializationError, UnknownRecordTypeError

__all__ = [
    'RecordSerializer',
    'ContextRecordSerializer',
    'DestinationRecordSerializer',
    'UnknownRecordTypeError',
    'SerializationError',
]



# Copyright (c) Qotto, 2021

"""
Messaging errors
"""

__all__ = [
    'MessagingError',
    'ConsumerCreationError',
    'ConsumerPollError',
    'ConsumerCommitError',
    'ProducerProduceError',
    'ProducerCreationError',
    'ProducerTransactionError',
]


class MessagingError(BaseException):
    """
    Base Error class of all messaging API
    """


class ConsumerCreationError(MessagingError):
    """
    A consumer could not be initialized
    """


class ConsumerPollError(MessagingError):
    """
    A consumer could not poll messages
    """


class ConsumerCommitError(MessagingError):
    """
    A consumer could not commit messages
    """


class ProducerCreationError(MessagingError):
    """
    A producer could not be initialized
    """


class ProducerProduceError(MessagingError):
    """
    A producer could not produce a message
    """


class ProducerTransactionError(MessagingError):
    """
    A transactional producer or processor could not commit a transaction
    """



# Copyright (c) Qotto, 2021

"""
Record errors
"""

from typing import Any, Union

__all__ = [
    'RecordAttributeError',
    'RecordAttributeTypeError',
    'RecordAttributeValueError',
]


class RecordAttributeError(Exception):
    """
    Base class for Record related errors
    """


class RecordAttributeTypeError(RecordAttributeError):
    """
    Record attribute has wrong type
    """

    def __init__(self, attribute: str, expected_type: Union[str, type], value: Any) -> None:
        super().__init__(
            f'Record attribute "{attribute}" should be of type "{expected_type}", '
            f'value was "{value}" of type {type(value)}'
        )


class RecordAttributeValueError(RecordAttributeError):
    """
    Record attribute has wrong value
    """

    def __init__(self, attribute: str, value: Any, message: str = None) -> None:
        super().__init__(
            f'Record attribute "{attribute}" has incorrect value "{value}" of type {type(value)}'
            f' (cause: {message or "unspecified"})'
        )



# Copyright (c) Qotto, 2021

"""
Kafka messaging errors
"""

from eventy.messaging.messaging_errors import (
    MessagingError,
    ConsumerPollError,
    ProducerTransactionError,
    ConsumerCommitError,
    ProducerProduceError,
    ConsumerCreationError,
    ProducerCreationError,
)

__all__ = [
    'KafkaError',
    'KafkaConsumerCreationError',
    'KafkaConsumerPollError',
    'KafkaTopicDeletionError',
    'KafkaProducerProduceError',
    'KafkaProducerTransactionError',
    'KafkaConsumerCommitError',
    'KafkaProducerCreationError',
    'KafkaTopicCreationError',
]


class KafkaError(MessagingError):
    """
    Base for all Kafka related errors
    """


class KafkaConsumerCreationError(KafkaError, ConsumerCreationError):
    """
    A Kafka consumer could not be created
    """


class KafkaConsumerPollError(KafkaError, ConsumerPollError):
    """
    A Kafka consumer could not poll messages
    """


class KafkaConsumerCommitError(KafkaError, ConsumerCommitError):
    """
    A Kafka consumer could not commit offsets
    """


class KafkaProducerCreationError(KafkaError, ProducerCreationError):
    """
    A Kafka producer could not be created
    """


class KafkaProducerProduceError(KafkaError, ProducerProduceError):
    """
    A kafka producer could not produce a message
    """


class KafkaProducerTransactionError(KafkaError, ProducerTransactionError):
    """
    A kafka producer could not commit a transaction
    """


class KafkaTopicCreationError(KafkaError):
    """
    Topic could not be created
    """


class KafkaTopicDeletionError(KafkaError):
    """
    Topic could not be deleted
    """
